{-# OPTIONS_GHC -cpp -XMagicHash #-}
{- For Hugs, use the option -F"cpp -P -traditional" -}

module LinearScan.Verify where


import Debug.Trace (trace, traceShow)
import qualified Prelude
import qualified Data.IntMap
import qualified Data.IntSet
import qualified Data.List
import qualified Data.Ord
import qualified Data.Functor.Identity
import qualified Hask.Utils

import qualified LinearScan.Applicative as Applicative
import qualified LinearScan.Blocks as Blocks
import qualified LinearScan.Class as Class
import qualified LinearScan.Contravariant as Contravariant
import qualified LinearScan.Functor as Functor
import qualified LinearScan.IntMap as IntMap
import qualified LinearScan.IntSet as IntSet
import qualified LinearScan.Lens as Lens
import qualified LinearScan.List1 as List1
import qualified LinearScan.Loops as Loops
import qualified LinearScan.Monad as Monad
import qualified LinearScan.Prelude0 as Prelude0
import qualified LinearScan.Resolve as Resolve
import qualified LinearScan.Ssr as Ssr
import qualified LinearScan.State0 as State0
import qualified LinearScan.UsePos as UsePos
import qualified LinearScan.Vector0 as Vector0
import qualified LinearScan.Eqtype as Eqtype
import qualified LinearScan.Fintype as Fintype
import qualified LinearScan.Seq as Seq
import qualified LinearScan.Ssrnat as Ssrnat



#ifdef __GLASGOW_HASKELL__
import qualified GHC.Base as GHC.Base
import qualified GHC.Prim as GHC.Prim
#else
-- HUGS
import qualified LinearScan.IOExts as IOExts
#endif


#ifdef __GLASGOW_HASKELL__
--unsafeCoerce :: a -> b
unsafeCoerce = GHC.Base.unsafeCoerce#
#else
-- HUGS
--unsafeCoerce :: a -> b
unsafeCoerce = IOExts.unsafeCoerce
#endif

type PhysReg = Prelude.Int

data UseVerifier =
   VerifyDisabled
 | VerifyEnabled

type RegAllocations =
  Vector0.Vec ((,) (Prelude.Maybe Blocks.VarId) (Prelude.Maybe Blocks.VarId))

data RegStateDesc =
   Build_RegStateDesc RegAllocations IntSet.IntSet

rsAllocs :: Prelude.Int -> RegStateDesc -> RegAllocations
rsAllocs maxReg r =
  case r of {
   Build_RegStateDesc rsAllocs0 rsStack0 -> rsAllocs0}

rsStack :: Prelude.Int -> RegStateDesc -> IntSet.IntSet
rsStack maxReg r =
  case r of {
   Build_RegStateDesc rsAllocs0 rsStack0 -> rsStack0}

residency :: (Functor.Functor a1) -> ((Prelude.Maybe Blocks.VarId) -> a1) ->
             ((,) (Prelude.Maybe Blocks.VarId) (Prelude.Maybe Blocks.VarId))
             -> a1
residency h x x0 =
  Lens._1 h x x0

reservation :: (Functor.Functor a1) -> ((Prelude.Maybe Blocks.VarId) -> a1)
               -> ((,) (Prelude.Maybe Blocks.VarId)
               (Prelude.Maybe Blocks.VarId)) -> a1
reservation h x x0 =
  Lens._2 h x x0

newRegStateDesc :: Prelude.Int -> RegStateDesc
newRegStateDesc maxReg =
  Build_RegStateDesc
    (Vector0.vconst maxReg ((,) Prelude.Nothing Prelude.Nothing))
    IntSet.emptyIntSet

data AllocError =
   VarNotAllocated Blocks.VarId
 | VarNotResident Blocks.VarId
 | VarNotResidentForReg Blocks.VarId Prelude.Int (Prelude.Maybe Blocks.VarId)
 | VarNotReservedForReg Blocks.VarId Prelude.Int (Prelude.Maybe Blocks.VarId)
 | PhysRegAlreadyResidentForVar Prelude.Int Blocks.VarId
 | PhysRegAlreadyReservedForVar Prelude.Int Blocks.VarId
 | RegAlreadyReservedToVar Prelude.Int Blocks.VarId Blocks.VarId
 | BlockWithoutPredecessors Blocks.BlockId
 | UnknownPredecessorBlock Blocks.BlockId Blocks.BlockId
 | ErrorAtBlockEnd Blocks.BlockId

type RegStateSig = RegStateDesc

packRegState :: Prelude.Int -> RegStateDesc -> RegStateDesc
packRegState maxReg rd =
  rd

data VerifiedSig a =
   Build_VerifiedSig RegStateDesc (IntMap.IntMap RegStateSig) (IntMap.IntMap
                                                              ([]
                                                              Resolve.ResolvingMoveSet)) 
 (IntMap.IntMap ([] AllocError)) a

verDesc :: Prelude.Int -> (VerifiedSig a1) -> RegStateDesc
verDesc maxReg v =
  case v of {
   Build_VerifiedSig verDesc0 verBlocks0 verMoves0 verErrors0 verExt0 ->
    verDesc0}

verBlocks :: Prelude.Int -> (VerifiedSig a1) -> IntMap.IntMap RegStateSig
verBlocks maxReg v =
  case v of {
   Build_VerifiedSig verDesc0 verBlocks0 verMoves0 verErrors0 verExt0 ->
    verBlocks0}

verMoves :: Prelude.Int -> (VerifiedSig a1) -> IntMap.IntMap
            ([] Resolve.ResolvingMoveSet)
verMoves maxReg v =
  case v of {
   Build_VerifiedSig verDesc0 verBlocks0 verMoves0 verErrors0 verExt0 ->
    verMoves0}

verErrors :: Prelude.Int -> (VerifiedSig a1) -> IntMap.IntMap ([] AllocError)
verErrors maxReg v =
  case v of {
   Build_VerifiedSig verDesc0 verBlocks0 verMoves0 verErrors0 verExt0 ->
    verErrors0}

verExt :: Prelude.Int -> (VerifiedSig a1) -> a1
verExt maxReg v =
  case v of {
   Build_VerifiedSig verDesc0 verBlocks0 verMoves0 verErrors0 verExt0 ->
    verExt0}

newVerifiedSig :: Prelude.Int -> a1 -> VerifiedSig a1
newVerifiedSig maxReg i =
  Build_VerifiedSig (newRegStateDesc maxReg) IntMap.emptyIntMap
    IntMap.emptyIntMap IntMap.emptyIntMap i

_verDesc :: Prelude.Int -> (Functor.Functor a2) ->
            (Contravariant.Contravariant a2) -> (RegStateDesc -> a2) ->
            (VerifiedSig a1) -> a2
_verDesc maxReg h h0 f s =
  Functor.fmap h (Prelude0.const s) (f (verDesc maxReg s))

_verState :: Prelude.Int -> (Functor.Functor a2) -> (RegStateDesc -> a2) ->
             (VerifiedSig a1) -> a2
_verState maxReg h f s =
  Functor.fmap h (\x -> Build_VerifiedSig ( x) (verBlocks maxReg s)
    (verMoves maxReg s) (verErrors maxReg s) (verExt maxReg s))
    (f (verDesc maxReg s))

_verBlocks :: Prelude.Int -> (Functor.Functor a2) -> ((IntMap.IntMap
              RegStateSig) -> a2) -> (VerifiedSig a1) -> a2
_verBlocks maxReg h f s =
  Functor.fmap h (\x -> Build_VerifiedSig (verDesc maxReg s) x
    (verMoves maxReg s) (verErrors maxReg s) (verExt maxReg s))
    (f (verBlocks maxReg s))

_verMoves :: Prelude.Int -> (Functor.Functor a2) -> ((IntMap.IntMap
             ([] Resolve.ResolvingMoveSet)) -> a2) -> (VerifiedSig a1) -> a2
_verMoves maxReg h f s =
  Functor.fmap h (\x -> Build_VerifiedSig (verDesc maxReg s)
    (verBlocks maxReg s) x (verErrors maxReg s) (verExt maxReg s))
    (f (verMoves maxReg s))

_verErrors :: Prelude.Int -> (Functor.Functor a2) -> ((IntMap.IntMap
              ([] AllocError)) -> a2) -> (VerifiedSig a1) -> a2
_verErrors maxReg h f s =
  Functor.fmap h (\x -> Build_VerifiedSig (verDesc maxReg s)
    (verBlocks maxReg s) (verMoves maxReg s) x (verExt maxReg s))
    (f (verErrors maxReg s))

_verExt :: Prelude.Int -> (Functor.Functor a2) -> (a1 -> a2) -> (VerifiedSig
           a1) -> a2
_verExt maxReg h f s =
  Functor.fmap h (\x -> Build_VerifiedSig (verDesc maxReg s)
    (verBlocks maxReg s) (verMoves maxReg s) (verErrors maxReg s) x)
    (f (verExt maxReg s))

type Verified mType a a0 = State0.StateT (VerifiedSig a) mType a0

errorsT :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId -> ([] AllocError)
           -> Verified a1 a2 ()
errorsT maxReg mDict pc errs =
  Lens.applyStateT (\_ -> _verErrors maxReg)
    (IntMap.coq_IntMap_insert pc errs) mDict

errorT :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId -> AllocError ->
          Verified a1 a2 ()
errorT maxReg mDict pc err =
  errorsT maxReg mDict pc ((:) err [])

addMove :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId ->
           Resolve.ResolvingMoveSet -> Verified a1 a2 ()
addMove maxReg mDict pc mv =
  Lens.applyStateT (\_ -> _verMoves maxReg)
    (IntMap.coq_IntMap_alter (\mxs -> Prelude.Just
      (case mxs of {
        Prelude.Just xs -> Seq.rcons xs mv;
        Prelude.Nothing -> (:) mv []})) pc) mDict

reserveReg :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId -> PhysReg ->
              Blocks.VarId -> Verified a1 a2 ()
reserveReg maxReg mDict pc reg var =
  Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
    Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
      case Ssr.prop
             (Eqtype.eq_op (Eqtype.option_eqType Ssrnat.nat_eqType)
               (Lens.view
                 (Lens.stepdownl' (unsafeCoerce (\_ -> reservation)))
                 (Vector0.vnth maxReg (rsAllocs maxReg st) reg))
               (unsafeCoerce Prelude.Nothing)) of {
       Prelude.Just _ ->
        Lens.modifyStateT (\_ -> _verState maxReg)
          (packRegState maxReg (Build_RegStateDesc
            (Vector0.vmodify maxReg (rsAllocs maxReg st) reg
              (Lens.set (\_ -> reservation) (Prelude.Just var)))
            (rsStack maxReg st))) mDict;
       Prelude.Nothing ->
        case Lens.view (Lens.stepdownl' (\_ -> reservation))
               (Vector0.vnth maxReg (rsAllocs maxReg st) reg) of {
         Prelude.Just v ->
          Monad.when (State0.coq_StateT_Monad mDict)
            (Prelude.not
              (Eqtype.eq_op Ssrnat.nat_eqType (unsafeCoerce v)
                (unsafeCoerce var)))
            (errorT maxReg mDict pc (RegAlreadyReservedToVar ( reg) v var));
         Prelude.Nothing ->
          Applicative.pure (State0.coq_StateT_Applicative mDict) ()}})
      (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict))
    (addMove maxReg mDict pc (Resolve.RSAllocReg var ( reg)))

isReserved :: Prelude.Int -> (Monad.Monad a1) -> PhysReg -> Blocks.VarId ->
              Verified a1 a2 Prelude.Bool
isReserved maxReg mDict reg var =
  Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
    Applicative.pure (State0.coq_StateT_Applicative mDict)
      (Eqtype.eq_op (Eqtype.option_eqType Ssrnat.nat_eqType)
        (Lens.view (Lens.stepdownl' (unsafeCoerce (\_ -> reservation)))
          (Vector0.vnth maxReg (rsAllocs maxReg st) reg))
        (unsafeCoerce (Prelude.Just var))))
    (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict)

checkReservation :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId -> PhysReg
                    -> Blocks.VarId -> Verified a1 a2 ()
checkReservation maxReg mDict pc reg var =
  Monad.bind (State0.coq_StateT_Monad mDict) (\b ->
    Monad.unless (State0.coq_StateT_Monad mDict) b
      (Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
        errorT maxReg mDict pc (VarNotReservedForReg var ( reg)
          (Lens.view (Lens.stepdownl' (\_ -> reservation))
            (Vector0.vnth maxReg (rsAllocs maxReg st) reg))))
        (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict)))
    (isReserved maxReg mDict reg var)

releaseReg :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId -> PhysReg ->
              Blocks.VarId -> Verified a1 a2 ()
releaseReg maxReg mDict pc reg var =
  Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
    Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
      case Ssr.prop
             (Eqtype.eq_op (Eqtype.option_eqType Ssrnat.nat_eqType)
               (Lens.view
                 (Lens.stepdownl' (unsafeCoerce (\_ -> reservation)))
                 (Vector0.vnth maxReg (rsAllocs maxReg st) reg))
               (unsafeCoerce (Prelude.Just var))) of {
       Prelude.Just _ ->
        Lens.modifyStateT (\_ -> _verState maxReg)
          (packRegState maxReg (Build_RegStateDesc
            (Vector0.vmodify maxReg (rsAllocs maxReg st) reg
              (Lens.set (\_ -> reservation) Prelude.Nothing))
            (rsStack maxReg st))) mDict;
       Prelude.Nothing ->
        errorT maxReg mDict pc (VarNotReservedForReg var ( reg)
          (Lens.view (Lens.stepdownl' (\_ -> reservation))
            (Vector0.vnth maxReg (rsAllocs maxReg st) reg)))})
      (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict))
    (addMove maxReg mDict pc (Resolve.RSFreeReg ( reg) var))

assignReg :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId -> PhysReg ->
             Blocks.VarId -> Verified a1 a2 ()
assignReg maxReg mDict pc reg var =
  Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
    Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
      case Ssr.prop
             (Eqtype.eq_op (Eqtype.option_eqType Ssrnat.nat_eqType)
               (Lens.view
                 (Lens.stepdownl' (unsafeCoerce (\_ -> reservation)))
                 (Vector0.vnth maxReg (rsAllocs maxReg st) reg))
               (unsafeCoerce (Prelude.Just var))) of {
       Prelude.Just _ ->
        Lens.modifyStateT (\_ -> _verState maxReg)
          (packRegState maxReg (Build_RegStateDesc
            (Vector0.vmodify maxReg (rsAllocs maxReg st) reg
              (Lens.set (\_ -> residency) (Prelude.Just var)))
            (rsStack maxReg st))) mDict;
       Prelude.Nothing ->
        errorT maxReg mDict pc (VarNotReservedForReg ( reg) var
          (Lens.view (Lens.stepdownl' (\_ -> reservation))
            (Vector0.vnth maxReg (rsAllocs maxReg st) reg)))})
      (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict))
    (addMove maxReg mDict pc (Resolve.RSAssignReg var ( reg)))

isResident :: Prelude.Int -> (Monad.Monad a1) -> PhysReg -> Blocks.VarId ->
              Verified a1 a2 Prelude.Bool
isResident maxReg mDict reg var =
  Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
    Applicative.pure (State0.coq_StateT_Applicative mDict)
      (Eqtype.eq_op (Eqtype.option_eqType Ssrnat.nat_eqType)
        (Lens.view (Lens.stepdownl' (unsafeCoerce (\_ -> residency)))
          (Vector0.vnth maxReg (rsAllocs maxReg st) reg))
        (unsafeCoerce (Prelude.Just var))))
    (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict)

checkResidency :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId -> PhysReg
                  -> Blocks.VarId -> Verified a1 a2 ()
checkResidency maxReg mDict pc reg var =
  Monad.bind (State0.coq_StateT_Monad mDict) (\b ->
    Monad.unless (State0.coq_StateT_Monad mDict) b
      (Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
        errorT maxReg mDict pc (VarNotResidentForReg var ( reg)
          (Lens.view (Lens.stepdownl' (\_ -> residency))
            (Vector0.vnth maxReg (rsAllocs maxReg st) reg))))
        (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict)))
    (isResident maxReg mDict reg var)

checkLiveness :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId ->
                 IntSet.IntSet -> Verified a1 a2 ()
checkLiveness maxReg mDict pc vars =
  Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
    Monad.forM_ (State0.coq_StateT_Monad mDict)
      (IntSet.coq_IntSet_toList vars) (\var ->
      Monad.unless (State0.coq_StateT_Monad mDict)
        (Vector0.vfoldl_with_index maxReg (\reg b p ->
          (Prelude.||) b
            (case Lens.view (Lens.stepdownl' (\_ -> residency)) p of {
              Prelude.Just var0 -> Prelude.True;
              Prelude.Nothing -> Prelude.False})) Prelude.False
          (rsAllocs maxReg st)) (errorT maxReg mDict pc (VarNotResident var))))
    (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict)

verifyBlockBegin :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId ->
                    UseVerifier -> Prelude.Int -> IntSet.IntSet ->
                    Loops.LoopState -> Verified a1 a2 ()
verifyBlockBegin maxReg mDict pc useVerifier bid liveIns loops =
  case useVerifier of {
   VerifyDisabled ->
    Applicative.pure (State0.coq_StateT_Applicative mDict) ();
   VerifyEnabled ->
    Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
      checkLiveness maxReg mDict pc liveIns)
      (case IntMap.coq_IntMap_lookup bid (Loops.forwardBranches loops) of {
        Prelude.Just fwds ->
         Monad.forM_ (State0.coq_StateT_Monad mDict)
           (IntSet.coq_IntSet_toList fwds) (\pred ->
           Monad.bind (State0.coq_StateT_Monad mDict) (\exits ->
             case IntMap.coq_IntMap_lookup pred exits of {
              Prelude.Just allocs ->
               Lens.modifyStateT (\_ -> _verState maxReg) allocs mDict;
              Prelude.Nothing ->
               errorT maxReg mDict pc (UnknownPredecessorBlock bid pred)})
             (Lens.use (Lens.stepdownl' (\_ -> _verBlocks maxReg)) mDict));
        Prelude.Nothing ->
         Monad.when (State0.coq_StateT_Monad mDict)
           ((Prelude.<=) ((Prelude.succ) 0) (IntSet.coq_IntSet_size liveIns))
           (errorT maxReg mDict pc (BlockWithoutPredecessors bid))})}

verifyBlockEnd :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId ->
                  UseVerifier -> Prelude.Int -> IntSet.IntSet -> Verified 
                  a1 a2 ()
verifyBlockEnd maxReg mDict pc useVerifier bid liveOuts =
  case useVerifier of {
   VerifyDisabled ->
    Applicative.pure (State0.coq_StateT_Applicative mDict) ();
   VerifyEnabled ->
    Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
      Monad.bind (State0.coq_StateT_Monad mDict) (\allocs ->
        Monad.bind (State0.coq_StateT_Monad mDict) (\x0 ->
          Lens.applyStateT (\_ -> _verBlocks maxReg)
            (IntMap.coq_IntMap_insert bid allocs) mDict)
          (Lens.modifyStateT (\_ -> _verState maxReg)
            (packRegState maxReg (newRegStateDesc maxReg)) mDict))
        (Lens.use (Lens.stepdownl' (\_ -> _verState maxReg)) mDict))
      (checkLiveness maxReg mDict pc liveOuts)}

verifyApplyAllocs :: Prelude.Int -> (Monad.Monad a3) -> (Blocks.OpInfo 
                     a3 a1 a2) -> Blocks.OpId -> UseVerifier -> a1 -> ([]
                     ((,) Blocks.VarId PhysReg)) -> Verified a3 a4 ([] a2)
verifyApplyAllocs maxReg mDict oinfo pc useVerifier op allocs =
  Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
    Class.lift (unsafeCoerce State0.coq_StateT_MonadTrans) mDict
      (State0.coq_StateT_Monad mDict)
      (Blocks.applyAllocs maxReg mDict oinfo op allocs))
    (case useVerifier of {
      VerifyDisabled ->
       Applicative.pure (State0.coq_StateT_Applicative mDict) ();
      VerifyEnabled ->
       Monad.forM_ (State0.coq_StateT_Monad mDict)
         (List1.sortBy (\x y ->
           UsePos.coq_VarKind_leq (Blocks.varKind maxReg x)
             (Blocks.varKind maxReg y))
           (Blocks.opRefs maxReg mDict oinfo op)) (\ref ->
         case Blocks.varId maxReg ref of {
          Prelude.Left reg ->
           Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
             case Blocks.varKind maxReg ref of {
              UsePos.Input ->
               Applicative.pure (State0.coq_StateT_Applicative mDict) ();
              UsePos.Temp ->
               case Vector0.vnth maxReg (rsAllocs maxReg st) reg of {
                (,) o o0 ->
                 case o0 of {
                  Prelude.Just v ->
                   errorT maxReg mDict pc (PhysRegAlreadyReservedForVar
                     ( reg) v);
                  Prelude.Nothing ->
                   Applicative.pure (State0.coq_StateT_Applicative mDict) ()}};
              UsePos.Output ->
               case Vector0.vnth maxReg (rsAllocs maxReg st) reg of {
                (,) o o0 ->
                 case o0 of {
                  Prelude.Just v ->
                   errorT maxReg mDict pc (PhysRegAlreadyReservedForVar
                     ( reg) v);
                  Prelude.Nothing ->
                   Applicative.pure (State0.coq_StateT_Applicative mDict) ()}}})
             (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict);
          Prelude.Right var ->
           case List1.maybeLookup Ssrnat.nat_eqType (unsafeCoerce allocs)
                  (unsafeCoerce var) of {
            Prelude.Just reg ->
             case Blocks.varKind maxReg ref of {
              UsePos.Input -> checkResidency maxReg mDict pc reg var;
              UsePos.Temp -> checkReservation maxReg mDict pc reg var;
              UsePos.Output ->
               Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
                 assignReg maxReg mDict pc reg var)
                 (checkReservation maxReg mDict pc reg var)};
            Prelude.Nothing -> errorT maxReg mDict pc (VarNotAllocated var)}})})

verifyResolutions :: Prelude.Int -> (Monad.Monad a1) -> Blocks.OpId ->
                     UseVerifier -> ([] Resolve.ResGraphEdge) -> Verified 
                     a1 a2 ([] Resolve.ResolvingMove)
verifyResolutions maxReg mDict pc useVerifier moves =
  case useVerifier of {
   VerifyDisabled ->
    Applicative.pure (State0.coq_StateT_Applicative mDict)
      (Prelude.map (Resolve.resMove maxReg) moves);
   VerifyEnabled ->
    Functor.fmap
      (State0.coq_StateT_Functor
        (Applicative.is_functor (Monad.is_applicative mDict))) Seq.rev
      (Monad.forFoldM (State0.coq_StateT_Monad mDict) [] moves (\acc mv ->
        Monad.bind (State0.coq_StateT_Monad mDict) (\st ->
          case Resolve.resMove maxReg mv of {
           Resolve.Move fromReg fromVar toReg ->
            Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
              Monad.bind (State0.coq_StateT_Monad mDict) (\x0 ->
                Monad.bind (State0.coq_StateT_Monad mDict) (\x1 ->
                  Monad.bind (State0.coq_StateT_Monad mDict) (\x2 ->
                    Monad.bind (State0.coq_StateT_Monad mDict) (\x3 ->
                      Monad.bind (State0.coq_StateT_Monad mDict) (\x4 ->
                        Applicative.pure
                          (State0.coq_StateT_Applicative mDict) ((:)
                          (Resolve.resMove maxReg mv) acc))
                        (Monad.when (State0.coq_StateT_Monad mDict)
                          (Resolve.resGhost maxReg mv)
                          (releaseReg maxReg mDict pc toReg fromVar)))
                      (assignReg maxReg mDict pc toReg fromVar))
                    (addMove maxReg mDict pc
                      (Resolve.weakenResolvingMove maxReg
                        (Resolve.resMove maxReg mv))))
                  (reserveReg maxReg mDict pc toReg fromVar))
                (Monad.unless (State0.coq_StateT_Monad mDict)
                  (Eqtype.eq_op (Fintype.ordinal_eqType maxReg)
                    (unsafeCoerce fromReg) (unsafeCoerce toReg))
                  (releaseReg maxReg mDict pc fromReg fromVar)))
              (checkResidency maxReg mDict pc fromReg fromVar);
           Resolve.Swap fromReg fromVar toReg toVar ->
            Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
              Monad.bind (State0.coq_StateT_Monad mDict) (\x0 ->
                Monad.bind (State0.coq_StateT_Monad mDict) (\x1 ->
                  Monad.bind (State0.coq_StateT_Monad mDict) (\x2 ->
                    Monad.bind (State0.coq_StateT_Monad mDict) (\x3 ->
                      Monad.bind (State0.coq_StateT_Monad mDict) (\x4 ->
                        Monad.bind (State0.coq_StateT_Monad mDict) (\x5 ->
                          Monad.bind (State0.coq_StateT_Monad mDict) (\x6 ->
                            Monad.bind (State0.coq_StateT_Monad mDict)
                              (\x7 ->
                              Applicative.pure
                                (State0.coq_StateT_Applicative mDict) ((:)
                                (Resolve.resMove maxReg mv) acc))
                              (assignReg maxReg mDict pc toReg fromVar))
                            (assignReg maxReg mDict pc fromReg toVar))
                          (reserveReg maxReg mDict pc toReg fromVar))
                        (reserveReg maxReg mDict pc fromReg toVar))
                      (addMove maxReg mDict pc
                        (Resolve.weakenResolvingMove maxReg
                          (Resolve.resMove maxReg mv))))
                    (releaseReg maxReg mDict pc toReg toVar))
                  (releaseReg maxReg mDict pc fromReg fromVar))
                (checkResidency maxReg mDict pc toReg toVar))
              (checkResidency maxReg mDict pc fromReg fromVar);
           Resolve.Spill fromReg toSpillSlot ->
            Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
              Monad.bind (State0.coq_StateT_Monad mDict) (\check ->
                case check of {
                 Prelude.True ->
                  Monad.bind (State0.coq_StateT_Monad mDict) (\x0 ->
                    Applicative.pure (State0.coq_StateT_Applicative mDict)
                      ((:) (Resolve.resMove maxReg mv) acc))
                    (addMove maxReg mDict pc
                      (Resolve.weakenResolvingMove maxReg
                        (Resolve.resMove maxReg mv)));
                 Prelude.False ->
                  Applicative.pure (State0.coq_StateT_Applicative mDict) acc})
                (isResident maxReg mDict fromReg toSpillSlot))
              (releaseReg maxReg mDict pc fromReg toSpillSlot);
           Resolve.Restore fromSpillSlot toReg ->
            Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
              Monad.bind (State0.coq_StateT_Monad mDict) (\x0 ->
                Monad.bind (State0.coq_StateT_Monad mDict) (\x1 ->
                  Monad.bind (State0.coq_StateT_Monad mDict) (\x2 ->
                    Applicative.pure (State0.coq_StateT_Applicative mDict)
                      ((:) (Resolve.resMove maxReg mv) acc))
                    (Monad.when (State0.coq_StateT_Monad mDict)
                      (Resolve.resGhost maxReg mv)
                      (releaseReg maxReg mDict pc toReg fromSpillSlot)))
                  (assignReg maxReg mDict pc toReg fromSpillSlot))
                (addMove maxReg mDict pc
                  (Resolve.weakenResolvingMove maxReg
                    (Resolve.resMove maxReg mv))))
              (reserveReg maxReg mDict pc toReg fromSpillSlot);
           Resolve.AllocReg toVar toReg ->
            Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
              Monad.bind (State0.coq_StateT_Monad mDict) (\x0 ->
                Applicative.pure (State0.coq_StateT_Applicative mDict) acc)
                (Monad.when (State0.coq_StateT_Monad mDict)
                  (Resolve.resGhost maxReg mv)
                  (releaseReg maxReg mDict pc toReg toVar)))
              (reserveReg maxReg mDict pc toReg toVar);
           Resolve.FreeReg fromReg fromVar ->
            Monad.bind (State0.coq_StateT_Monad mDict) (\x ->
              Applicative.pure (State0.coq_StateT_Applicative mDict) acc)
              (releaseReg maxReg mDict pc fromReg fromVar)})
          (Lens.use (Lens.stepdowng (\_ -> _verDesc maxReg)) mDict)))}

